package ubiik.wp.ami.controller;

import java.io.File;
import java.io.FileInputStream;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.util.Calendar;
import java.util.Collection;
import java.util.LinkedList;
import java.util.Queue;
import java.util.concurrent.LinkedBlockingQueue;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.atomic.AtomicBoolean;
import java.util.stream.Collectors;
import javafx.application.Platform;
import javafx.fxml.FXML;
import javafx.scene.control.CheckBox;
import javafx.scene.control.ComboBox;
import javafx.scene.control.TextArea;
import javafx.stage.FileChooser;
import javafx.stage.Stage;
import org.controlsfx.control.textfield.TextFields;
import org.json.simple.JSONArray;
import org.json.simple.JSONObject;
import org.json.simple.parser.JSONParser;
import ubiik.utils.DatatypeConverter;
import ubiik.utils.JsonUtils;
import ubiik.wp.ami.model.AMIMessage;
import ubiik.wp.ami.model.AmiMessageFactory;
import ubiik.wp.ami.model.AmiPSCommand;
import ubiik.wp.controller.DevicesController;
import ubiik.wp.controller.MessageController;
import ubiik.wp.demo.controller.UplinksDemoViewController;
import ubiik.wp.model.devices.DevicesStorage;
import ubiik.wp.model.devices.UUEndDevice;
import ubiik.wp.model.messages.PSDownlink;
import ubiik.wp.model.messages.PSMessage;
import ubiik.wp.model.messages.PSUplink;
import ubiik.wp.model.messages.pscommands.PSCommand;
import ubiik.wp.view.DisplayConstants;

public class FANTestController extends UplinksDemoViewController implements Runnable {
  private Queue<AMITest> amiTests;
  
  private LinkedBlockingQueue<AMIMessage> amiMsgQueue;
  
  private AtomicBoolean runningTests;
  
  private JSONObject jsonTests;
  
  private Stage mainStage;
  
  @FXML
  protected ComboBox<String> comboEUUID;
  
  @FXML
  protected TextArea textArea;
  
  @FXML
  protected CheckBox cbAck;
  
  public FANTestController(DevicesController devicesController, MessageController messageController, Stage mainStage) {
    super(devicesController, messageController);
    this.amiMsgQueue = new LinkedBlockingQueue<>();
    this.amiTests = new LinkedList<>();
    this.runningTests = new AtomicBoolean(false);
    this.mainStage = mainStage;
  }
  
  protected void processUplink(PSUplink uplink) {
    System.out.println("Received " + DatatypeConverter.printHexBinary(uplink.getApplicationData()));
    AMIMessage msg = AmiMessageFactory.createAmiMessage(uplink);
    if (msg != null)
      try {
        this.amiMsgQueue.put(msg);
      } catch (InterruptedException e) {
        e.printStackTrace();
      }  
  }
  
  public AMIMessage getAMIMessage(long millisTimeout) {
    AMIMessage msg = null;
    try {
      msg = this.amiMsgQueue.poll(millisTimeout, TimeUnit.MILLISECONDS);
    } catch (InterruptedException e) {
      e.printStackTrace();
    } 
    return msg;
  }
  
  public AMIMessage getAMIMessage(long millisTimeout, byte msgType, Byte subtype) {
    long timeStart = System.currentTimeMillis();
    long remaining = millisTimeout;
    AMIMessage msg = null;
    boolean found = false;
    while (remaining > 0L && !found) {
      msg = getAMIMessage(remaining);
      remaining = millisTimeout - System.currentTimeMillis() - timeStart;
      if (msg != null) {
        found = (msg.getMessageId() == msgType);
        if (subtype != null)
          found = (found && subtype.equals(msg.getMessageSubType())); 
      } 
    } 
    return msg;
  }
  
  public void doReloadView() {
    super.doReloadView();
    if (this.comboEUUID.getItems().size() != DevicesStorage.getInstance().getDevices().size())
      setupIdInput(); 
  }
  
  private void setupIdInput() {
    Platform.runLater(new Runnable() {
          public void run() {
            FANTestController.this.comboEUUID.getItems().clear();
            FANTestController.this.comboEUUID.getItems().addAll((Collection)DevicesStorage.getInstance().getDevices().stream().map(d -> d.toString()).collect(Collectors.toList()));
            FANTestController.this.comboEUUID.setEditable(true);
            TextFields.bindAutoCompletion(FANTestController.this.comboEUUID.getEditor(), FANTestController.this.comboEUUID.getItems());
            if (FANTestController.this.comboEUUID.getItems().size() > 0)
              FANTestController.this.comboEUUID.getSelectionModel().select(0); 
          }
        });
  }
  
  public synchronized void btnStartPressed() {
    boolean wasRunning = this.runningTests.getAndSet(true);
    if (!wasRunning) {
      Thread threadRt = new Thread(this);
      threadRt.start();
    } 
  }
  
  public synchronized void btnLoadTestsPressed() {
    FileChooser fileChooser = new FileChooser();
    fileChooser.setTitle("Open Multicast Groups File");
    File selectedFile = fileChooser.showOpenDialog(this.mainStage);
    if (selectedFile != null)
      try {
        InputStream is = new FileInputStream(selectedFile);
        JSONParser parser = new JSONParser();
        this.jsonTests = (JSONObject)parser.parse(new InputStreamReader(is, "UTF-8"));
        loadTests(this.jsonTests);
      } catch (Exception e) {
        this.jsonTests = null;
        e.printStackTrace();
        StringBuilder errMsg = new StringBuilder("Make sure you are trying to open a valid AMI Test File.");
        if (e.getMessage() != null) {
          errMsg.append("\nError: ");
          errMsg.append(e.getMessage());
        } 
        displayErrorMessage("Error parsing file", errMsg.toString());
      }  
  }
  
  public void btnStopPressed() {
    this.runningTests.set(false);
  }
  
  public void btnClearTerminalPressed() {
    this.textArea.clear();
  }
  
  public void loadTests(JSONObject json) {
    this.amiTests.clear();
    JSONArray tests = (JSONArray)json.get("tests");
    int endpoint = this.cbAck.isSelected() ? 1 : 0;
    for (int i = 0; i < tests.size(); i++) {
      AMITest amiTest = null;
      String testType = (String)JsonUtils.getIfExists((JSONObject)tests.get(i), "type");
      if (testType != null)
        if (testType.compareTo("time") == 0) {
          amiTest = new AmiTestTime(this, (JSONObject)tests.get(i), endpoint);
        } else if (testType.compareTo("xanLog") == 0) {
          amiTest = new AmiXanLogTest(this, (JSONObject)tests.get(i), endpoint);
        } else if (testType.compareTo("stress") == 0) {
          amiTest = new AmiParallelStressTest(this, (JSONObject)tests.get(i), endpoint);
        }  
      if (amiTest == null)
        amiTest = new AMITest(this, (JSONObject)tests.get(i), endpoint); 
      this.amiTests.add(amiTest);
    } 
  }
  
  public void run() {
    if (this.jsonTests == null) {
      displayErrorMessage("Select a test file", "Please load a test file");
    } else {
      appendToTextArea(this.textArea, "------ Starting tests ------\n");
      String strDevice = this.comboEUUID.getSelectionModel().getSelectedItem();
      this.amiMsgQueue.clear();
      System.out.println("Selected device: " + strDevice.toString());
      byte[] device = DatatypeConverter.parseHexBinary(strDevice);
      if (device != null && device.length == 16) {
        loadTests(this.jsonTests);
        while (!this.amiTests.isEmpty() && this.runningTests.get()) {
          AMITest amiTest = this.amiTests.poll();
          amiTest.setFan(device);
          appendToTextArea(this.textArea, DisplayConstants.SIMPLE_DATE_FORMAT_TIME_DATE
              .format(Calendar.getInstance().getTime()) + " Running test: " + DisplayConstants.SIMPLE_DATE_FORMAT_TIME_DATE.format(Calendar.getInstance().getTime()) + "\n");
          UUTestResult result = amiTest.run();
          appendToTextArea(this.textArea, result.getMessage());
          appendToTextArea(this.textArea, result.isSuccess() ? "OK" : "Failed");
          appendToTextArea(this.textArea, "\n\n");
        } 
        appendToTextArea(this.textArea, "------ Finished running tests ------\n");
      } else {
        displayErrorMessage("No device selected", "Please select a device");
      } 
    } 
    this.runningTests.set(false);
  }
  
  public void sendAmiMessage(AMIMessage amiMessage, byte[] uuid, int endpoint) {
    if (amiMessage instanceof AmiPSCommand) {
      PSCommand psCmd = ((AmiPSCommand)amiMessage).getPsCommand();
      if (psCmd != null) {
        System.out.println("Send PSCommand");
        this.messageController.sendMessage((PSMessage)psCmd);
      } 
    } else if (amiMessage instanceof ubiik.wp.ami.model.AmiBSConnect) {
      if (!this.messageController.isConnected().get())
        Platform.runLater(new Runnable() {
              public void run() {
                FANTestController.this.connectionButtonPressed();
              }
            }); 
    } else if (amiMessage.getPayload() != null) {
      System.out.println("Send AMI message: " + DatatypeConverter.printHexBinary(amiMessage.getPayload()));
      this.messageController.sendMessage((PSMessage)new PSDownlink(amiMessage.getPayload(), uuid, endpoint));
    } else {
      System.out.println("Skip message with no payload (type " + amiMessage.getMessageId() + ")");
    } 
  }
  
  public boolean getBSInfo() {return false;}
}
