package ubiik.wp.ami.model;

import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.security.InvalidParameterException;
import java.util.Arrays;
import java.util.HashMap;
import java.util.Map;
import org.json.simple.JSONObject;
import ubiik.utils.DatatypeConverter;
import ubiik.utils.JsonUtils;
import ubiik.wp.model.messages.PSUplink;

public abstract class AMIMessage {
  public static final int HEADER_SIZE = 2;
  
  public static final int DEFAULT_TIMEOUT_MILLIS = 30000;
  
  private byte applicationId;
  
  private byte messageId;
  
  public enum Direction {
    in, out;
  }
  
  protected Byte messageSubType = null;
  
  protected ByteArrayOutputStream payload;
  
  protected Direction direction;
  
  private long timeout;
  
  private PSUplink uplink;
  
  public static final byte APP_ID = 20;
  
  public static final byte AmiMeterKeys = 0;
  
  public static final byte AmiMeterWrappedKeys = 1;
  
  public static final byte AmiGet = 2;
  
  public static final byte AmiSet = 3;
  
  public static final byte AmiAction = 4;
  
  public static final byte AmiGlobalGet = 5;
  
  public static final byte AmiGlobalSet = 6;
  
  public static final byte AmiGlobalAction = 7;
  
  public static final byte AmiRawDlms = 16;
  
  public static final byte AmiRawHdlc = 17;
  
  public static final byte AmiGetMeterDetails = 32;
  
  public static final byte AmiGetXanLog = 33;
  
  public static final byte AmiGetFanParameters = 48;
  
  public static final byte AmiSetFanParameters = 49;
  
  public static final byte AmiGetHanParameters = 50;
  
  public static final byte AmiSetHanParameters = 51;
  
  public static final byte AmiMeterFirmwareUpdate = 64;
  
  public static final byte AmiSetMeterTime = 65;
  
  public static final byte AmiLoadProfile = -128;
  
  public static final byte AmiRawAxdr = -127;
  
  public static final byte AmiPackAxdr = -126;
  
  public static final byte AmiEvent = -125;
  
  public static final byte AmiError = -124;
  
  public static final byte AmiGetPackedResponse = -119;
  
  public static final byte AmiGetResponse = -118;
  
  public static final byte AmiSetResponse = -117;
  
  public static final byte AmiActionResponse = -116;
  
  public static final byte AmiPong = -115;
  
  public static final byte AmiRawDlmsResponse = -112;
  
  public static final byte AmiRawHdlcResponse = -111;
  
  public static final byte AmiMeterDiscovery = -96;
  
  public static final byte AmiNotification = -95;
  
  public static final byte AmiMeterDetails = -94;
  
  public static final byte AmiXanLog = -93;
  
  public static final byte AmiFanParameters = -92;
  
  public static final byte AmiHanParameters = -91;
  
  public static final byte AmiMeterFirmwareUpdateStatus = -80;
  
  public static final byte AmiUnknown = -1;
  
  public static final byte AmiAtCommand = -2;
  
  public static final byte AmiCustom = -3;
  
  public static final byte AmiSleep = -4;
  
  public static final byte AmiPSCommand = -5;
  
  public static final byte AmiBSConnect = -6;
  
  public static final byte AmiRandomPayload = -7;
  
  public static final byte AmiXanFan = 0;
  
  public static final byte AmiXanHan = 1;
  
  public static final byte AmiAuthenticated = 0;
  
  public static final byte AmiNotAuthenticated = 1;
  
  public static final byte AmiMeterNotFound = 2;
  
  public static final String JSON_KEY_MESSAGES = "messages";
  
  public static final String JSON_KEY_AMI_ERR_CODE = "errorCode";
  
  public static final String JSON_KEY_DIRECTION = "dir";
  
  public static final String JSON_KEY_DIR_IN = "in";
  
  public static final String JSON_KEY_DIR_OUT = "out";
  
  public static final String JSON_KEY_MSG_TYPE = "type";
  
  public static final String JSON_KEY_NOTIFICATION_TYPE = "notification";
  
  public static final String JSON_KEY_ASSOC_STATUS = "associationStatus";
  
  public static final String JSON_KEY_TIMEOUT = "timeout";
  
  public static final String JSON_KEY_METER_UNIQUE_ID = "meterUniqueId";
  
  public static final String JSON_KEY_METER_ID = "meterId";
  
  public static final String JSON_KEY_METER_TYPE = "meterType";
  
  public static final String JSON_KEY_GUK = "guk";
  
  public static final String JSON_KEY_AK = "ak";
  
  public static final String JSON_KEY_DEVICE_NAME = "deviceName";
  
  public static final String JSON_KEY_DISTRICT_ID = "districtId";
  
  public static final String JSON_KEY_CUSTOMER_ID = "customerId";
  
  public static final String JSON_KEY_METER_TYPE_CODE = "meterTypeCode";
  
  public static final String JSON_KEY_FW_INFO = "firmwareVersion";
  
  public static final String JSON_KEY_METER_DETAILS_X = "x";
  
  public static final String JSON_KEY_METER_DETAILS_Y = "y";
  
  public static final String JSON_KEY_METER_DETAILS_FAN_TYPE = "fanType";
  
  public static final String JSON_KEY_METER_DETAILS_HAN_TYPE = "hanType";
  
  public static final String JSON_KEY_FAN_PARAM_OPT = "opt";
  
  public static final String JSON_KEY_FAN_PARAM_POLL_PRD = "pollPeriodSec";
  
  public static final String JSON_KEY_FAN_PARAM_CONNECT_PRD = "connectPeriodSec";
  
  public static final String JSON_KEY_FAN_PARAM_SET_TIME_PRD = "setTimePeriod";
  
  public static final String JSON_KEY_FAN_PARAM_TIME_TOL = "timeTolerance";
  
  public static final String JSON_KEY_FAN_PARAM_INIT_DELAY = "initialDelay";
  
  public static final String JSON_KEY_HAN_QOS = "qos";
  
  public static final String JSON_KEY_HAN_REQUEST = "request";
  
  public static final String JSON_KEY_TO = "to";
  
  public static final String JSON_KEY_FROM = "from";
  
  public static final String JSON_KEY_TO_BYTES = "toBytes";
  
  public static final String JSON_KEY_FROM_BYTES = "fromBytes";
  
  public static final String JSON_KEY_BLOCKED_FRAMES = "blockedFrames";
  
  public static final String JSON_KEY_HDLC_ERRORS = "hdlcErrors";
  
  public static final String JSON_KEY_XAN_TYPE = "xanType";
  
  public static final String JSON_KEY_XAN_ID = "xanId";
  
  public static final String JSON_KEY_REQ_ID = "requestId";
  
  public static final String JSON_KEY_CLASS_ID = "classId";
  
  public static final String JSON_KEY_OBIS_CODE = "obisCode";
  
  public static final String JSON_KEY_ATTR_ID = "attributeId";
  
  public static final String JSON_KEY_SELECT_TYPE = "selectType";
  
  public static final String JSON_KEY_REQ_BLOCK = "reqBlock";
  
  public static final String JSON_KEY_STATUS = "status";
  
  public static final String JSON_KEY_AXDR = "axdr";
  
  public static final String JSON_KEY_REG = "reg";
  
  public static final String JSON_KEY_START = "start";
  
  public static final String JSON_KEY_END = "end";
  
  public static final String JSON_KEY_HDLC_PDU = "hdlcPdu";
  
  public static final String JSON_KEY_RECORD_NUMBER = "recordNumber";
  
  public static final String JSON_KEY_AMI_TIME = "time";
  
  public static final String JSON_KEY_VALUE = "value";
  
  public static final String JSON_KEY_VALUES = "values";
  
  public static final String JSON_KEY_AT_COMMAND = "atCommand";
  
  public static final String JSON_KEY_AMI_PAYLOAD = "amiPayload";
  
  public static final String JSON_KEY_CUSTOM_MESSAGE_ID = "messageId";
  
  public static final String JSON_KEY_SLEEP = "sleep";
  
  public static final String JSON_KEY_SIZE = "size";
  
  public static final String JSON_KEY_MILLIS = "millis";
  
  public static final String JSON_AmiMeterKeys = "meterKeys";
  
  public static final String JSON_AmiMeterWrappedKeys = "wrappedKeys";
  
  public static final String JSON_AmiGet = "get";
  
  public static final String JSON_AmiSet = "set";
  
  public static final String JSON_AmiAction = "action";
  
  public static final String JSON_AmiGlobalGet = "globalGet";
  
  public static final String JSON_AmiGlobalSet = "globalSet";
  
  public static final String JSON_AmiGlobalAction = "globalAction";
  
  public static final String JSON_AmiRawDlms = "rawDlms";
  
  public static final String JSON_AmiRawHdlc = "rawHdlc";
  
  public static final String JSON_AmiGetMeterDetails = "getMeterDetails";
  
  public static final String JSON_AmiGetXanLog = "getXanLog";
  
  public static final String JSON_AmiGetFanParameters = "getFanParameters";
  
  public static final String JSON_AmiSetFanParameters = "setFanParameters";
  
  public static final String JSON_AmiGetHanParameters = "getHanParameters";
  
  public static final String JSON_AmiSetHanParameters = "setHanParameters";
  
  public static final String JSON_AmiMeterFirmwareUpdate = "meterFirmwareUpdate";
  
  public static final String JSON_AmiSetMeterTime = "setMeterTime";
  
  public static final String JSON_AmiLoadProfile = "loadProfile";
  
  public static final String JSON_AmiRawAxdr = "rawAxdr";
  
  public static final String JSON_AmiPackAxdr = "packAxdr";
  
  public static final String JSON_AmiEvent = "event";
  
  public static final String JSON_AmiGetPackedResponse = "getPackedResponse";
  
  public static final String JSON_AmiGetResponse = "getResponse";
  
  public static final String JSON_AmiSetResponse = "setResponse";
  
  public static final String JSON_AmiActionResponse = "actionResponse";
  
  public static final String JSON_AmiPong = "pong";
  
  public static final String JSON_AmiRawDlmsResponse = "rawDlmsResponse";
  
  public static final String JSON_AmiRawHdlcResponse = "rawHdlcResponse";
  
  public static final String JSON_AmiMeterDiscovery = "meterDiscovery";
  
  public static final String JSON_AmiNotification = "notification";
  
  public static final String JSON_AmiMeterDetails = "meterDetails";
  
  public static final String JSON_AmiXanLog = "xanLog";
  
  public static final String JSON_AmiFanParameters = "fanParameters";
  
  public static final String JSON_AmiHanParameters = "hanParameters";
  
  public static final String JSON_AmiMeterFirmwareUpdateStatus = "meterFirmwareUpdateStatus";
  
  public static final String JSON_AmiUnknown = "unknown";
  
  public static final String JSON_AmiError = "amiError";
  
  public static final String JSON_AmiCustom = "custom";
  
  public static final String JSON_AmiSleep = "sleep";
  
  public static final String JSON_AmiPSCommand = "psCommand";
  
  public static final String JSON_AmiBSConnect = "bsConnect";
  
  public static final String JSON_AmiRandomPayload = "randomPayload";
  
  public static final String JSON_KEY_COMMAND = "cmd";
  
  public AMIMessage(PSUplink uplink) {
    this(uplink.getApplicationData());
    this.uplink = uplink;
  }
  
  public AMIMessage(byte[] bytes, Direction dir) {
    if (bytes.length < 2)
      throw new IllegalArgumentException("Data is too small"); 
    this.direction = dir;
    this.applicationId = bytes[0];
    this.messageId = bytes[1];
    this.payload = new ByteArrayOutputStream();
    this.timeout = 30000L;
    try {
      this.payload.write(bytes);
    } catch (IOException e) {
      e.printStackTrace();
      throw new IllegalArgumentException("Error writing data");
    } 
  }
  
  public AMIMessage(byte[] bytes) {
    this(bytes, Direction.in);
  }
  
  public AMIMessage(byte messageId, Direction dir) {
    this.uplink = null;
    this.timeout = 30000L;
    this.direction = dir;
    this.applicationId = 20;
    this.messageId = messageId;
    this.payload = new ByteArrayOutputStream();
    this.payload.write(this.applicationId);
    this.payload.write(this.messageId);
  }
  
  public AMIMessage(JSONObject json) {
    this(getMessaIdForKey((String)json.get("type")), 
        getDirectionForKey((String)json.get("dir")));
    Long longTout = JsonUtils.getLongIfExists(json, "timeout");
    if (longTout != null)
      this.timeout = longTout.longValue(); 
  }
  
  public static final Direction getDirectionForKey(String key) {
    if (key.compareTo("in") == 0)
      return Direction.in; 
    if (key.compareTo("out") == 0)
      return Direction.out; 
    throw new InvalidParameterException("Invalid direction");
  }
  
  public static final byte getMessaIdForKey(String key) {
    Byte value = AmiMsgs.get(key);
    if (value == null)
      throw new InvalidParameterException("Invalid Message ID: " + key); 
    return value.byteValue();
  }
  
  protected abstract AmiComparisonResult doAmiCompare(AMIMessage paramAMIMessage);
  
  public AmiComparisonResult amiCompare(AMIMessage other) {
    if (getClass().equals(other.getClass()))
      return doAmiCompare(other); 
    return new AmiComparisonResult(false, "Message types don't match");
  }
  
  public static Boolean compareIfNotNull(byte[] obj, byte[] other, Boolean result, StringBuilder msg, String error) {
    if (obj != null && other != null) {
      Boolean b = Boolean.valueOf((Arrays.compare(obj, other) == 0));
      if (!b.booleanValue() && msg != null && error != null) {
        msg.append(error);
        msg.append("Expected ");
        msg.append(DatatypeConverter.printHexBinary(obj));
        msg.append(" but received ");
        msg.append(DatatypeConverter.printHexBinary(other));
        msg.append("\n");
      } 
      result = Boolean.valueOf(result.booleanValue() & b.booleanValue());
    } 
    return result;
  }
  
  public static Boolean compareIfNotNull(Object obj, Object other, Boolean result, StringBuilder msg, String error) {
    if (obj != null && other != null) {
      Boolean b = Boolean.valueOf(obj.equals(other));
      if (!b.booleanValue() && msg != null && error != null) {
        msg.append(error);
        msg.append("Expected ");
        msg.append(obj.toString());
        msg.append(" but received ");
        msg.append(other.toString());
        msg.append("\n");
      } 
      result = Boolean.valueOf(result.booleanValue() & b.booleanValue());
    } 
    return result;
  }
  
  public byte getApplicationId() {
    return this.applicationId;
  }
  
  public byte getMessageId() {
    return this.messageId;
  }
  
  public byte[] getPayload() {
    return this.payload.toByteArray();
  }
  
  public Direction getDirection() {
    return this.direction;
  }
  
  public long getTimeout() {
    return this.timeout;
  }
  
  public PSUplink getUplink() {
    return this.uplink;
  }
  
  public void setUplink(PSUplink ul) {
    this.uplink = ul;
  }
  
  public Byte getMessageSubType() {
    return this.messageSubType;
  }
  
  private static final Map<String, Byte> AmiMsgs = new HashMap<>();
  
  static {
    AmiMsgs.put("meterKeys", Byte.valueOf((byte)0));
    AmiMsgs.put("wrappedKeys", Byte.valueOf((byte)1));
    AmiMsgs.put("get", Byte.valueOf((byte)2));
    AmiMsgs.put("set", Byte.valueOf((byte)3));
    AmiMsgs.put("action", Byte.valueOf((byte)4));
    AmiMsgs.put("globalGet", Byte.valueOf((byte)5));
    AmiMsgs.put("globalSet", Byte.valueOf((byte)6));
    AmiMsgs.put("globalAction", Byte.valueOf((byte)7));
    AmiMsgs.put("rawDlms", Byte.valueOf((byte)16));
    AmiMsgs.put("rawHdlc", Byte.valueOf((byte)17));
    AmiMsgs.put("getMeterDetails", Byte.valueOf((byte)32));
    AmiMsgs.put("getXanLog", Byte.valueOf((byte)33));
    AmiMsgs.put("getFanParameters", Byte.valueOf((byte)48));
    AmiMsgs.put("setFanParameters", Byte.valueOf((byte)49));
    AmiMsgs.put("getHanParameters", Byte.valueOf((byte)50));
    AmiMsgs.put("setHanParameters", Byte.valueOf((byte)51));
    AmiMsgs.put("meterFirmwareUpdate", Byte.valueOf((byte)64));
    AmiMsgs.put("setMeterTime", Byte.valueOf((byte)65));
    AmiMsgs.put("loadProfile", Byte.valueOf((byte)-128));
    AmiMsgs.put("rawAxdr", Byte.valueOf((byte)-127));
    AmiMsgs.put("packAxdr", Byte.valueOf((byte)-126));
    AmiMsgs.put("event", Byte.valueOf((byte)-125));
    AmiMsgs.put("getPackedResponse", Byte.valueOf((byte)-119));
    AmiMsgs.put("getResponse", Byte.valueOf((byte)-118));
    AmiMsgs.put("setResponse", Byte.valueOf((byte)-117));
    AmiMsgs.put("actionResponse", Byte.valueOf((byte)-116));
    AmiMsgs.put("pong", Byte.valueOf((byte)-115));
    AmiMsgs.put("rawDlmsResponse", Byte.valueOf((byte)-112));
    AmiMsgs.put("rawHdlcResponse", Byte.valueOf((byte)-111));
    AmiMsgs.put("meterDiscovery", Byte.valueOf((byte)-96));
    AmiMsgs.put("notification", Byte.valueOf((byte)-95));
    AmiMsgs.put("meterDetails", Byte.valueOf((byte)-94));
    AmiMsgs.put("xanLog", Byte.valueOf((byte)-93));
    AmiMsgs.put("fanParameters", Byte.valueOf((byte)-92));
    AmiMsgs.put("hanParameters", Byte.valueOf((byte)-91));
    AmiMsgs.put("meterFirmwareUpdateStatus", Byte.valueOf((byte)-80));
    AmiMsgs.put("unknown", Byte.valueOf((byte)-1));
    AmiMsgs.put("amiError", Byte.valueOf((byte)-124));
    AmiMsgs.put("custom", Byte.valueOf((byte)-3));
    AmiMsgs.put("sleep", Byte.valueOf((byte)-4));
    AmiMsgs.put("psCommand", Byte.valueOf((byte)-5));
    AmiMsgs.put("bsConnect", Byte.valueOf((byte)-6));
    AmiMsgs.put("randomPayload", Byte.valueOf((byte)-7));
  }
}
