package ubiik.wp.controller;

import java.nio.ByteBuffer;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Date;
import java.util.List;
import java.util.Map;
import java.util.concurrent.atomic.AtomicBoolean;
import ubiik.utils.DatatypeConverter;
import ubiik.wp.model.bsparameters.BSParameter;
import ubiik.wp.model.devices.DevicesStorage;
import ubiik.wp.model.devices.MulticastGroup;
import ubiik.wp.model.devices.UUEndDevice;
import ubiik.wp.model.devices.UUSIB0;
import ubiik.wp.model.fpga.FPGARegister;
import ubiik.wp.model.messages.PSDownlink;
import ubiik.wp.model.messages.PSMessage;
import ubiik.wp.model.messages.PSMessageDeviceStatus;
import ubiik.wp.model.messages.PSPong;
import ubiik.wp.model.messages.PSUplink;
import ubiik.wp.model.messages.pscommands.PSCommandPingEndDevice;
import ubiik.wp.model.sdr.SDRParameter;

public class DevicesController {
  private DevicesStorage devicesStorage = DevicesStorage.getInstance();
  
  private InactiveDevicesManager inactiveDevicesManager = null;
  
  private AtomicBoolean enableInactiveDevicesMonitoring = new AtomicBoolean(false);
  
  private static final boolean DEFAULT_ENABLE_INACTIVE_DEVICES_MONITORING = false;
  
  private PingManager pingManager = new PingManager(this);
  
  public void setIdxForEds(Map<String, Integer> edsIndx, int startIdx, boolean createDevices) {
    this.devicesStorage.setIdxForEds(edsIndx, startIdx);
    if (createDevices)
      for (String uueidStr : edsIndx.keySet())
        this.devicesStorage.addDevice(DatatypeConverter.parseHexBinary(uueidStr));  
  }
  
  public UUEndDevice addDevice(byte[] uuid) {
    return this.devicesStorage.addDevice(uuid);
  }
  
  public void lookForNewDeviceInMessage(PSMessage message) {
    byte[] target = null;
    Long totalUlACK = null;
    Long totalUlUNACK = null;
    if (message instanceof PSUplink) {
      PSUplink uplink = (PSUplink)message;
      target = uplink.getTarget();
    } else if (message instanceof PSMessageDeviceStatus) {
      PSMessageDeviceStatus deviceStatus = (PSMessageDeviceStatus)message;
      if (deviceStatus.isForEndDevice()) {
        target = deviceStatus.getDeviceId();
        totalUlACK = deviceStatus.getTotalULACK();
        totalUlUNACK = deviceStatus.getTotalULUNACK();
      } 
    } 
    if (target != null && 
      !this.devicesStorage.deviceExists(target)) {
      this.devicesStorage.addDevice(target);
      UUEndDevice ed = this.devicesStorage.getDeviceForId(target);
      if (ed != null && 
        totalUlACK != null && totalUlUNACK != null && ed != null)
        ed.setUplinksCounter(totalUlACK.longValue() + totalUlUNACK.longValue()); 
    } 
  }
  
  public ArrayList<FPGARegister> getFpgaRegisters() {
    return this.devicesStorage.getBaseStation().getRegisters();
  }
  
  public boolean setAllFPGARegisters(byte[] data) {
    boolean success = false;
    if (data.length == 4 * getFpgaRegisters().size()) {
      ByteBuffer buff = ByteBuffer.wrap(data);
      ArrayList<FPGARegister> registers = getFpgaRegisters();
      for (int i = 0; i < registers.size(); i++)
        ((FPGARegister)registers.get(i)).setFrom32Bit(buff.getInt()); 
      success = true;
    } else {
      System.err.println("Discard command response. Received " + data.length + " bytes instead of " + 4 * getFpgaRegisters().size());
    } 
    return success;
  }
  
  public synchronized void restart(boolean reloadIndeces, boolean createDevices) {
    this.devicesStorage.restart();
    if (reloadIndeces) {
      EndDeviceNameAndIdxManager edIdxManager = EndDeviceNameAndIdxManager.getInstance();
      if (edIdxManager.getMaxIndex() >= 0)
        setIdxForEds(edIdxManager.getDevicesAndIndexes(), edIdxManager.getMaxIndex(), createDevices); 
    } 
  }
  
  public int getLastChannelSet() {
    return this.devicesStorage.getBaseStation().getChannel();
  }
  
  public void setBaseFrequency(int channel) {
    this.devicesStorage.getBaseStation().setChannel(channel);
  }
  
  public List<Integer> getBlacklistedChannels() {
    return this.devicesStorage.getBaseStation().getBlacklistedChannels();
  }
  
  public boolean addBlacklistedChannel(int channel) {
    return this.devicesStorage.getBaseStation().addBlacklistedChannel(channel);
  }
  
  public void removeBlacklistedChannel(int channel) {
    this.devicesStorage.getBaseStation().removeBlacklistedChannel(channel);
  }
  
  public void clearBlacklistedChannels() {
    this.devicesStorage.getBaseStation().clearBlacklistedChannels();
  }
  
  public boolean isChannelBlacklisted(int channel) {
    return this.devicesStorage.getBaseStation().isChannelBlacklisted(channel);
  }
  
  public synchronized void addDevicesToMulticastGroup(int groupId, List<byte[]> devices, List<Byte> status) {
    this.devicesStorage.updateDevicesInMulticastGroup(groupId, devices, status);
  }
  
  public synchronized void removeDevicesFromMulticastGrouo(int groupId, List<byte[]> devices) {
    this.devicesStorage.removeDevicesFromMulticastGrouo(groupId, devices);
  }
  
  public synchronized Map<Integer, MulticastGroup> getMulticastGroups() {
    return this.devicesStorage.getMulticastGroups();
  }
  
  public synchronized boolean mutlicastGroupExists(int groupId) {
    return this.devicesStorage.mutlicastGroupExists(groupId);
  }
  
  public synchronized boolean addMulticastGroup(MulticastGroup mcg) {
    return this.devicesStorage.addMulticastGroup(mcg);
  }
  
  public UUSIB0 getSIB() {
    return this.devicesStorage.getSib();
  }
  
  public void setSIB(UUSIB0 sib) {
    this.devicesStorage.setSib(sib);
  }
  
  public void setBSParameterValue(int paramIdx, Long value) {
    this.devicesStorage.setBSRParameterValue(paramIdx, value);
  }
  
  public ArrayList<BSParameter> getBSParameters() {
    return this.devicesStorage.getBSParameters();
  }
  
  public BSParameter getBSParameterForIdx(int paramIdx) {
    return this.devicesStorage.getBSRParameterForIdx(paramIdx);
  }
  
  public void setSDRParameterValue(int sdrParaId, String value) {
    this.devicesStorage.setSDRParameterValue(sdrParaId, value);
  }
  
  public Collection<SDRParameter> getSDRParameters() {
    return this.devicesStorage.getSDRParameters();
  }
  
  public SDRParameter getSDRParameterForId(int paramId) {
    return this.devicesStorage.getSDRParameterForId(paramId);
  }
  
  public byte[] getBaseStationId() {
    return this.devicesStorage.getBsid();
  }
  
  public void setMulticastGroups(Map<Integer, List<byte[]>> multicastGroups, Map<Integer, List<Byte>> status) {
    this.devicesStorage.setMulticastGroups(multicastGroups, status);
  }
  
  public synchronized void addMulticastGroups(Map<Integer, List<byte[]>> multicastGroups, Map<Integer, List<Byte>> status) {
    this.devicesStorage.addMulticastGroups(multicastGroups, status);
  }
  
  public synchronized boolean setScheduleTransferForDevice(byte[] uuid, byte flag, int window, int interval) {
    return this.devicesStorage.setScheduleTransferForDevice(uuid, flag, window, interval);
  }
  
  public void setRunMode(Integer runMode) {
    this.devicesStorage.setRunMode(runMode);
  }
  
  public Integer getRunMode() {
    return this.devicesStorage.getRunMode();
  }
  
  public synchronized UUEndDevice getDeviceForId(byte[] uuid) {
    return this.devicesStorage.getDeviceForId(uuid);
  }
  
  public synchronized byte[] getBsVersion() {
    return this.devicesStorage.getBsVersion();
  }
  
  public synchronized Date getBsBuildDate() {
    return this.devicesStorage.getBsBuildDate();
  }
  
  public synchronized Collection<UUEndDevice> getDevices() {
    return this.devicesStorage.getDevices();
  }
  
  public synchronized void clearDevices() {
    this.devicesStorage.clearDevices();
  }
  
  public synchronized void startTrackingInactiveDevices(MessageController mc) {
    if (this.enableInactiveDevicesMonitoring.get()) {
      if (this.inactiveDevicesManager == null)
        this.inactiveDevicesManager = new InactiveDevicesManager(300000L, 60000L, this, mc); 
      if (!this.inactiveDevicesManager.isRunning()) {
        Thread t = new Thread(this.inactiveDevicesManager);
        t.start();
      } 
    } 
  }
  
  public synchronized void stopTrackingInactiveDevices() {
    if (this.inactiveDevicesManager != null && 
      this.inactiveDevicesManager.isRunning())
      this.inactiveDevicesManager.stop(); 
  }
  
  public synchronized void setEnableInactiveDevicesMonitoring(boolean enable) {
    if (!enable)
      stopTrackingInactiveDevices(); 
    this.enableInactiveDevicesMonitoring.set(enable);
  }
  
  public synchronized boolean isInactiveDevicesMonitoringEnabled() {
    return this.enableInactiveDevicesMonitoring.get();
  }
  
  public Long getInactiveDevicesPingTimeout() {
    if (this.inactiveDevicesManager != null)
      return Long.valueOf(this.inactiveDevicesManager.getPingTimeout()); 
    return null;
  }
  
  public boolean setInactiveDevicesPingTimeout(long pingTimeout) {
    if (this.inactiveDevicesManager != null) {
      this.inactiveDevicesManager.setPingTimeout(pingTimeout);
      return true;
    } 
    return false;
  }
  
  public Long getInactiveDevicesCheckPeriod() {
    if (this.inactiveDevicesManager != null)
      return Long.valueOf(this.inactiveDevicesManager.getCheckPeriod()); 
    return null;
  }
  
  public boolean setInactiveDevicesCheckPeriod(long checkPeriod) {
    if (this.inactiveDevicesManager != null) {
      this.inactiveDevicesManager.setCheckPeriod(checkPeriod);
      return true;
    } 
    return false;
  }
  
  public Long getInactiveDevicesInactivePeriod() {
    if (this.inactiveDevicesManager != null)
      return Long.valueOf(this.inactiveDevicesManager.getInactivePeriod()); 
    return null;
  }
  
  public boolean setInactiveDevicesInactivePeriod(long period) {
    if (this.inactiveDevicesManager != null) {
      this.inactiveDevicesManager.setInactivePeriod(period);
      return true;
    } 
    return false;
  }
  
  public synchronized PSPong getLastPongForDevice(byte[] uueid) {
    return this.devicesStorage.getLastPongForDevice(uueid);
  }
  
  public synchronized void setLastPongForDevice(PSPong lastPong) {
    this.devicesStorage.setLastPongForDevice(lastPong);
  }
  
  public PingManager getPingManager() {
    return this.pingManager;
  }
  
  public long getNextPingId() {
    return this.pingManager.getNextPingId();
  }
  
  public void addPong(PSPong pong) {
    this.pingManager.addPong(pong);
  }
  
  public void addPing(PSCommandPingEndDevice ping) {
    this.pingManager.addPing(ping);
  }
  
  public synchronized void addUplinkToDevice(PSUplink uplink) {
    this.devicesStorage.addUplinkToDevice(uplink);
  }
  
  public synchronized void addDownlinkToDevice(PSDownlink downlink) {
    this.devicesStorage.addDownlinkToDevice(downlink);
  }
  
  public synchronized void setBsid(byte[] bsid) {
    this.devicesStorage.setBsid(bsid);
  }
  
  
  public synchronized void setEndDeviceStatus(byte[] uuid, byte status, Date timeStamp, short rssi, Date lastActivity, Date dateConnected, Long totalUplinksUNACK, Long totalUplinksACK, byte deviceType, byte[] firmwareVersion) {
	    this.devicesStorage.setEndDeviceStatus(uuid, status, timeStamp, rssi, lastActivity, dateConnected, totalUplinksUNACK, totalUplinksACK, deviceType, firmwareVersion);
  }
  
  public synchronized void setEndDeviceStatus(byte[] uuid, byte status, Date timeStamp, short rssi, Date lastActivity, Date dateConnected, Long totalUplinksUNACK, Long totalUplinksACK, byte deviceType, byte[] firmwareVersion, Long joinFoi) {
    this.devicesStorage.setEndDeviceStatus(uuid, status, timeStamp, rssi, lastActivity, dateConnected, totalUplinksUNACK, totalUplinksACK, deviceType, firmwareVersion, joinFoi);
  }
  
  public synchronized void setBsVersionAndBuildDate(byte[] version, String strDate) {
    this.devicesStorage.setBsVersionAndBuildDate(version, strDate);
  }
  
  public synchronized int getNumEndDevices() {
	return this.devicesStorage.getNumEndDevices();
  }
  
  public synchronized void setNumEndDevices(int numEDs) {
	  this.devicesStorage.setNumEndDevices(numEDs);
  }
  
}
