package ubiik.wp.controller;

import java.util.LinkedList;
import java.util.List;
import java.util.concurrent.atomic.AtomicBoolean;
import ubiik.wp.model.messages.PSAckMessage;
import ubiik.wp.model.messages.PSDownlink;
import ubiik.wp.model.messages.PSMessage;
import ubiik.wp.psconnection.PSConnectionErrorCode;
import ubiik.wp.psconnection.PSConnectionException;
import ubiik.wp.psconnection.PSConnectionManager;
import ubiik.wp.psconnection.PSConnectionManagerDelegate;
import ubiik.wp.util.ApplicationLogger;

public class MessageController implements PSConnectionManagerDelegate, MessageReceiver {
  private List<ConnectionObserver> connectionObservers;
  
  private PSConnectionManager connectionManager;
  
  private IncomingMessagesProcessor msgProcessor;
  
  private Thread msgProcessorThread;
  
  private AtomicBoolean manuallyDisconnected;
  
  private DevicesController devicesController;
  
  public MessageController(DevicesController devicesController) {
    this.msgProcessorThread = null;
    this.connectionManager = null;
    this.msgProcessor = null;
    this.devicesController = devicesController;
    this.connectionObservers = new LinkedList<>();
    this.manuallyDisconnected = new AtomicBoolean(false);
  }
  
  public void connect(String host, int port, PSConnectionManager connManager) {
    if (this.msgProcessorThread == null) {
      this.manuallyDisconnected.set(false);
      this.connectionManager = connManager;
      this.msgProcessor = new IncomingMessagesProcessor(connManager.getInMessagesQueue(), this);
      this.msgProcessorThread = new Thread(this.msgProcessor);
      this.msgProcessorThread.start();
      try {
        this.connectionManager.connect();
      } catch (PSConnectionException e) {
        authenticationFinished(false);
      } 
    } 
  }
  
  public void disconnect() {
    this.manuallyDisconnected.set(true);
    this.connectionManager.disconnect();
    this.connectionManager = null;
    this.msgProcessorThread.interrupt();
    this.msgProcessorThread = null;
  }
  
  public synchronized void sendMessage(PSMessage message) {
    if (this.connectionManager != null) {
      ApplicationLogger.debug("Send message. Type: " + message.getTypeCode() + " seq #: " + message
          .getMessageId() + " data: " + message.getDataString());
      PSMessageStorage.getInstance().addMessage(message);
      if (message instanceof PSDownlink) {
        PSDownlink dl = (PSDownlink)message;
        if (dl.getBytesForTarget() == 16)
          dl.setTargetExists(Boolean.valueOf((this.devicesController.getDeviceForId(dl.getTarget()) != null))); 
      } 
      this.connectionManager.sendBytes(message.toBytes());
      if (!(message instanceof ubiik.wp.model.messages.PSMessageResponse))
        PSMessageStorage.getInstance().storeSentMessage(message); 
    } 
  }
  
  public synchronized void sendMessages(List<PSMessage> messages) {
    for (PSMessage m : messages)
      sendMessage(m); 
  }
  
  public synchronized void receiveMessage(PSMessage message) {
    if (!(message instanceof ubiik.wp.model.messages.PSMessageResponse) && 
      !(message instanceof ubiik.wp.model.messages.PSMessageDiscovery)) {
      PSAckMessage ack = new PSAckMessage(message.getMessageId());
      sendMessage((PSMessage)ack);
      this.devicesController.lookForNewDeviceInMessage(message);
    } 
    PSMessageStorage.getInstance().addMessage(message);
  }
  
  public void connectionFailed(PSConnectionErrorCode errorCode) {
    if (!this.manuallyDisconnected.get()) {
    	disconnect();	
      
      synchronized (this.connectionObservers) {
        for (ConnectionObserver connectionObserver : this.connectionObservers) {	
        	 connectionObserver.connectionFailed(errorCode);
        }
      }
    }
  } 
  
  public void authenticationFinished(boolean autheticated) {
    if (!autheticated) {
    	disconnect();	
    } 
    synchronized (this.connectionObservers) {
      for (ConnectionObserver connectionObserver : this.connectionObservers) {
    	  connectionObserver.authenticationFinished(autheticated); 
      }
    } 
  }
  
  public void addConnectionObserver(ConnectionObserver observer) {
    synchronized (this.connectionObservers) {
      this.connectionObservers.add(observer);
    } 
  }
  
  public void removeConnectionObserver(ConnectionObserver observer) {
    synchronized (this.connectionObservers) {
      this.connectionObservers.remove(observer);
    } 
  }
  
  public AtomicBoolean isConnected() {
    if (this.connectionManager != null)
      return this.connectionManager.isAuthenticated(); 
    return new AtomicBoolean(false);
  }
  
  public DevicesController getDevicesController() {
    return this.devicesController;
  }
  
  public void restartMessageStorage() {
    PSMessageStorage.getInstance().restart();
  }
  
  public String getCurrentHost() {
    String host = null;
    if (this.connectionManager != null)
      host = this.connectionManager.getHost(); 
    return host;
  }
  
  public boolean getBSInfo() {
	// do not need to impl here.
	  synchronized (this.connectionObservers) {
		  for (ConnectionObserver connectionObserver : this.connectionObservers) {
    	  return connectionObserver.getBSInfo(); 
      }
    }
	return false; 
  }
}
