package ubiik.wp.model.devices;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Date;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.stream.Collectors;
import ubiik.wp.model.bsparameters.BSParameter;
import ubiik.wp.model.fpga.FPGARegister;

public class UUBaseStation {
  public static final String PARAM_KEY_UT = "UT";
  
  public static final String PARAM_KEY_BS = "BS";
  
  private ArrayList<FPGARegister> registers;
  
  private ArrayList<BSParameter> bsParameters;
  
  private Object registersLock;
  
  private Object paramsLock;
  
  private byte[] bsid;
  
  private int channel;
  
  private byte[] version;
  
  private Date buildDate;
  
  private Integer runMode;
  
  private Integer numEndDevices;
  
  private String endDeviceType;
  
  public static final int DEFAULT_SUB_MODE = 0;
  
  private List<Integer> blackListedChannels;
  
  public UUBaseStation() {
    this.bsid = new byte[0];
    restart();
  }
  
  public void restart() {
    this.registers = null;
    this.registersLock = new Object();
    this.paramsLock = new Object();
    this.channel = -1;
    this.blackListedChannels = new LinkedList<>();
    this.bsParameters = null;
    this.runMode = null;
    this.numEndDevices = -1;
  }
  
  public void setRegisters(ArrayList<FPGARegister> reg) {
    synchronized (this.registersLock) {
      this.registers = reg;
    } 
  }
  
  public synchronized void setBsParameters(ArrayList<BSParameter> bsParameters) {
    synchronized (this.paramsLock) {
      this.bsParameters = bsParameters;
    } 
  }
  
  public synchronized ArrayList<BSParameter> getBsParameters() {
    synchronized (this.paramsLock) {
      return new ArrayList<>(this.bsParameters);
    } 
  }
  
  public synchronized ArrayList<BSParameter> getBsParameters(String category) {
    synchronized (this.paramsLock) {
      return new ArrayList<>((Collection<? extends BSParameter>)this.bsParameters.stream()
          .filter(elem -> (elem.getCategory().compareTo(category) == 0))
          .collect(Collectors.toList()));
    } 
  }
  
  public synchronized BSParameter getBsParameterAtIdx(int idx) {
    synchronized (this.paramsLock) {
      return this.bsParameters.get(idx);
    } 
  }
  
  public byte[] getBsid() {
    return this.bsid;
  }
  
  public void setBsid(byte[] bsid) {
    this.bsid = bsid;
  }
  
  public ArrayList<FPGARegister> getRegisters() {
    return new ArrayList<>(this.registers);
  }
  
  public boolean addBlacklistedChannel(int channel) {
    boolean success = true;
    synchronized (this.blackListedChannels) {
      if (this.blackListedChannels.indexOf(Integer.valueOf(channel)) < 0) {
        this.blackListedChannels.add(Integer.valueOf(channel));
      } else {
        success = false;
      } 
    } 
    return success;
  }
  
  public void removeBlacklistedChannel(int channel) {
    synchronized (this.blackListedChannels) {
      this.blackListedChannels.remove(Integer.valueOf(channel));
    } 
  }
  
  public void clearBlacklistedChannels() {
    synchronized (this.blackListedChannels) {
      this.blackListedChannels.clear();
    } 
  }
  
  public List<Integer> getBlacklistedChannels() {
    return new LinkedList<>(this.blackListedChannels);
  }
  
  public synchronized int getChannel() {
    return this.channel;
  }
  
  public synchronized void setChannel(int channel) {
    this.channel = channel;
  }
  
  public boolean isChannelBlacklisted(int channel) {
    return (this.blackListedChannels.indexOf(Integer.valueOf(channel)) >= 0);
  }
  
  public synchronized Integer getRunMode() {
    return this.runMode;
  }
  
  public synchronized void setRunMode(Integer runMode) {
    this.runMode = runMode;
  }
  
  private static List<Map<String, Integer>> getBaseChannelValidRanges() {
    List<Map<String, Integer>> validChannels = new ArrayList<>();
    HashMap<String, Integer> v1 = new HashMap<>();
    v1.put("min", Integer.valueOf(4000));
    v1.put("max", Integer.valueOf(5100));
    HashMap<String, Integer> v2 = new HashMap<>();
    v2.put("min", Integer.valueOf(8000));
    v2.put("max", Integer.valueOf(10200));
    validChannels.add(v1);
    validChannels.add(v2);
    return validChannels;
  }
  
  public static boolean validateBaseChannel(short channel) {
    boolean valid = false;
    int chn = 0xFFFF & channel;
    List<Map<String, Integer>> validValues = getBaseChannelValidRanges();
    for (Map<String, Integer> ranges : validValues) {
      Integer min = ranges.get("min");
      Integer max = ranges.get("max");
      if (chn >= min.intValue() && chn <= max.intValue()) {
        valid = true;
        break;
      } 
    } 
    return valid;
  }
  
  public synchronized byte[] getVersion() {
    return this.version;
  }
  
  public synchronized void setVersion(byte[] version) {
    this.version = version;
  }
  
  public synchronized Date getBuildDate() {
    return this.buildDate;
  }
  
  public synchronized void setBuildDate(Date buildDate) {
    this.buildDate = buildDate;
  }
  
  public synchronized int getNumEndDevices() {
	  return this.numEndDevices;
  }
  
  public synchronized void setNumEndDevices(int numEDs) {
	  this.numEndDevices = numEDs;
  }
}
