package ubiik.wp.psconnection;

import java.io.IOException;
import java.net.InetSocketAddress;
import java.net.Socket;
import java.util.Timer;
import java.util.TimerTask;
import java.util.concurrent.BlockingQueue;
import java.util.concurrent.atomic.AtomicBoolean;
import ubiik.wp.controller.UserManager;
import ubiik.wp.model.messages.PSMessage;
import ubiik.wp.model.messages.PSMessageAuthenticated;
import ubiik.wp.model.messages.PSMessageFactory;
import ubiik.wp.psconnection.auth.PSAuth;
import ubiik.wp.util.ApplicationLogger;

public class SocketPSConnectionManager extends PSConnectionManager {
  private AtomicBoolean receivedChallenge;
  
  private Socket socket;
  
  private PSInConnection inConnection;
  
  private PSOutConnection outConnection;
  
  private Thread threadInConnection;
  
  private Thread threadOutConnection;
  
  private Timer timeOutTimerTimer;
  
  private boolean gotBSInfo;
  
  private byte[] authenticate(byte[] challenge) {
    PSAuth psAuth = UserManager.getInstance().getPsAuth();
    return psAuth.encryptMessage(challenge);
  }
  
  private boolean processAuthResponse(byte[] response) {
    boolean accepted = false;
    try {
      PSMessage message = PSMessageFactory.createMessageFromBytes(response);
      if (message instanceof PSMessageAuthenticated) {
        accepted = ((PSMessageAuthenticated)message).isAuthenticated();
      } else {
        ApplicationLogger.debug("Expecting message type 0 but received " + message.getMessageId());
      } 
    } catch (Exception e) {
      accepted = false;
    } 
    return accepted;
  }
  
  public SocketPSConnectionManager(String host, int port, PSConnectionManagerDelegate delegate, BlockingQueue<byte[]> queue) {
    super(host, port, delegate, queue);
    this.socket = null;
    this.inConnection = null;
    this.outConnection = null;
    this.receivedChallenge = new AtomicBoolean(false);
    this.timeOutTimerTimer = null;
    this.gotBSInfo = false;
  }
  
  public void connect() throws PSConnectionException {
    if (this.socket == null) {
      try {
        this.socket = new Socket();
        this.socket.connect(new InetSocketAddress(this.host, this.port), 5000);
      } catch (IOException e) {
        throw new PSConnectionException();
      } 
      this.inConnection = new PSInConnection(this.socket, this);
      this.outConnection = new PSOutConnection(this.socket, this);
      this.threadInConnection = new Thread(this.inConnection);
      this.threadOutConnection = new Thread(this.outConnection);
      this.threadInConnection.start();
      this.threadOutConnection.start();
      final SocketPSConnectionManager connectionManager = this;
      TimerTask authTimeoutTask = new TimerTask() {
          public void run() {
            connectionManager.authenticationTimeFinished();
          }
        };
      if (this.timeOutTimerTimer != null)
        this.timeOutTimerTimer.cancel(); 
      this.timeOutTimerTimer = new Timer(true);
      this.timeOutTimerTimer.schedule(authTimeoutTask, 5000L);
    } 
  }
  
  public void disconnect() {
    if (this.timeOutTimerTimer != null)
      this.timeOutTimerTimer.cancel(); 
    if (this.threadInConnection != null) {
      this.threadInConnection.interrupt();
      this.threadInConnection = null;
    } 
    if (this.threadOutConnection != null) {
      this.threadOutConnection.interrupt();
      this.threadOutConnection = null;
    } 
    this.inConnection = null;
    this.outConnection = null;
    if (this.socket != null)
      try {
        this.socket.close();
        this.socket = null;
      } catch (IOException e) {
        e.printStackTrace();
        throw new PSConnectionException();
      }  
    this.authenticated.set(false);
    this.gotBSInfo = false;
  }
  
  public void authenticationTimeFinished() {
    this.timeOutTimerTimer = null;
    if (!isAuthenticated().get()) {
      ApplicationLogger.debug("Didn't receive authentication message. Discconnect");
      disconnect();
      connectionFailed(PSConnectionErrorCode.TIME_OUT);
    } 
  }
  
  public void sendBytes(byte[] bytes) {
    if (this.authenticated.get())
      this.outConnection.sendBytes(bytes); 
  }
  
  public AtomicBoolean isAuthenticated() {
    return this.authenticated;
  }
  
  public void receivedBytes(byte[] bytes) {
    if (!this.receivedBaseStationId.get()) {
      ApplicationLogger.debug("Received Base Station ID");
      this.receivedBaseStationId.set(true);
      try {
        this.inMessages.put(bytes);
      } catch (InterruptedException e) {
        e.printStackTrace();
      } 
    } else if (!this.authenticated.get()) {
      if (!this.receivedChallenge.get()) {
        byte[] ab = authenticate(bytes);
        this.outConnection.sendBytes(ab);
        this.receivedChallenge.set(true);
      } else {
        boolean accepted = processAuthResponse(bytes);
        if (accepted) {
          ApplicationLogger.debug("Authentication response: Accepted. Stay connected");
          this.authenticated.set(true);
        } else {
          ApplicationLogger.debug("Authentication response: Not accepted. Disconnect");
          disconnect();
        }
        this.delegate.authenticationFinished(accepted);
       
      } 
    } else {
      try {
    	if (!this.gotBSInfo) {
    		this.gotBSInfo = this.delegate.getBSInfo();
    	}
        this.inMessages.put(bytes);
      } catch (InterruptedException e) {
        ApplicationLogger.debug("PSConnectionManager: received interrupted");
      } 
    } 
  }
  
  public void connectionFailed(PSConnectionErrorCode errorCode) {
    if (this.delegate != null)
      this.delegate.connectionFailed(errorCode); 
    ApplicationLogger.debug("connection failed ERR: " + errorCode);
  }
}
