package ubiik.wp.view;

import java.util.Date;
import java.util.LinkedHashMap;
import java.util.Optional;
import javafx.collections.FXCollections;
import javafx.event.Event;
import javafx.event.EventHandler;
import javafx.fxml.FXML;
import javafx.scene.Cursor;
import javafx.scene.chart.LineChart;
import javafx.scene.chart.NumberAxis;
import javafx.scene.chart.XYChart;
import javafx.scene.control.Alert;
import javafx.scene.control.ButtonType;
import javafx.scene.control.Label;
import javafx.scene.control.ProgressBar;
import javafx.scene.control.TextField;
import javafx.scene.input.MouseEvent;
import javafx.scene.layout.Pane;
import javafx.scene.layout.StackPane;
import ubiik.utils.view.ViewUtils;
import ubiik.wp.controller.DevicesController;
import ubiik.wp.controller.MessageController;
import ubiik.wp.controller.PSMessageStorage;
import ubiik.wp.model.devices.UUBaseStation;
import ubiik.wp.model.frequencyscanner.FrequencyScanManager;
import ubiik.wp.model.frequencyscanner.FrequencyScanSample;
import ubiik.wp.model.messages.JSONMessageDelegate;
import ubiik.wp.model.messages.pscommands.PSCommand;
import ubiik.wp.model.messages.pscommands.PSCommandDelegate;
import ubiik.wp.model.messages.pscommands.PSCommandFrequencyScan;
import ubiik.wp.model.messages.pscommands.PSCommandRunMode;
import ubiik.wp.model.messages.pscommands.PSCommandSetBaseChannel;

public class FrequencyScannerViewController extends BaseViewController implements PSCommandDelegate {
  private static final int X_AXIS_TOTAL_TICKS = 10;
  
  private FrequencyScanManager frequencyScanManager;
  
  private Date lastUpdate;
  
  private static final String KEY_FREQ_SCAN = "frequency_scan";
  
  private static final short DEFAULT_TIME_WINDOW = 100;
  
  private static final int DEFAULT_MIN_X = 50;
  
  private static final int DEFAULT_MAX_Y = 0;
  
  private static final short INVALID_INPUT_SHORT = -1;
  
  private static final int MIN_WARFCN_STEP = 1;
  
  private static final int MAX_WARFCN_STEP = 10200;
  
  private static final int MIN_SCAN_TIME = 1;
  
  private static final int MAX_SCAN_TIME = 3600;
  
  @FXML
  private TextField inputStartWARFCN;
  
  @FXML
  private TextField inputStopWARFCN;
  
  @FXML
  private TextField inputStepWARFCN;
  
  @FXML
  private TextField inputScanTime;
  
  @FXML
  private TextField inputNewWARFCNValue;
  
  @FXML
  private Label labelScanStatus;
  
  @FXML
  private ProgressBar progressBar;
  
  @FXML
  protected Pane chartsContainer;
  
  @FXML
  private Label labelCurrentMark;
  
  private LineChart<Number, Number> warfcnVsRssiChart;
  
  private XYChart.Series<Number, Number> warfcnVsRssiSeries;
  
  private NumberAxis xAxis;
  
  private NumberAxis yAxis;
  
  private int minX;
  
  private int maxX;
  
  private int minY;
  
  private int maxY;
  
  private int step;
  
  public FrequencyScannerViewController(DevicesController devicesController, MessageController messageController, CommandStatusReceiver commandStatusReceiver) {
    super(devicesController, messageController, commandStatusReceiver);
    this.frequencyScanManager = new FrequencyScanManager();
    this.lastUpdate = null;
    this.warfcnVsRssiSeries = new XYChart.Series<>("WARFCN vs RSSI", FXCollections.observableArrayList());
    this.step = 10;
    restartAxisRanges();
  }
  
  public void initialize() {
    this.warfcnVsRssiChart = setUpLineChart(this.warfcnVsRssiChart, this.warfcnVsRssiSeries, "WARFCN vs RSSI");
    PSMessageStorage.getInstance().addJSONMessageDelegate("frequency_scan", (JSONMessageDelegate)this.frequencyScanManager);
  }
  
  private short validateChannel(TextField input) {
    boolean valid = true;
    short channel = -1;
    try {
      channel = Short.parseShort(input.getText());
    } catch (NumberFormatException e) {
      valid = false;
    } 
    valid = UUBaseStation.validateBaseChannel(channel);
    setTextFieldHasError(input, !valid);
    if (valid)
      return channel; 
    return -1;
  }
  
  public void btnScanPressed() {
    ViewUtils.updateProgressBar(this.progressBar, 0, 0);
    short startWARFCN = validateChannel(this.inputStartWARFCN);
    boolean valid = (startWARFCN != -1);
    short stopWARFCN = validateChannel(this.inputStopWARFCN);
    valid = (valid && stopWARFCN != -1);
    Integer stepWARFCN = validateAndGetIntegerTextFieldMinMax(this.inputStepWARFCN, 1, 10200);
    valid = (valid && stepWARFCN != null);
    Integer scanSeconds = validateAndGetIntegerTextFieldMinMax(this.inputScanTime, 1, 3600);
    valid = (valid && scanSeconds != null);
    if (valid) {
      restartAxisRanges();
      this.step = stepWARFCN.intValue();
      PSCommandFrequencyScan cmd = new PSCommandFrequencyScan(this, startWARFCN, stopWARFCN, stepWARFCN.shortValue(), (short)100, (short)(scanSeconds.intValue() * 1000 / 100));
      sendCommandAndUpdateStatusBar((PSCommand)cmd);
    } 
  }
  
  public void btnStopPressed() {
    sendCommandAndUpdateStatusBar((PSCommand)new PSCommandRunMode(this, (byte)1, 0));
  }
  
  public void btnClearPressed() {
    this.frequencyScanManager.clear();
  }
  
  @FXML
  public void btnSetBaseFrequencyPressed() {
    short channel = validateChannel(this.inputNewWARFCNValue);
    boolean valid = (channel != -1);
    if (valid) {
      Alert alert = new Alert(Alert.AlertType.CONFIRMATION);
      alert.setTitle("Set Base Channel");
      alert.setHeaderText("Setting the Base Channel will cause the Base Station to restart. All End Devices connected must also be restarted and set to the new channel.");
      alert.setContentText("Continue?");
      Optional<ButtonType> result = alert.showAndWait();
      if (result.isPresent() && result.get() == ButtonType.OK) {
        cleanInput(this.inputNewWARFCNValue);
        PSCommandSetBaseChannel cmd = new PSCommandSetBaseChannel(channel, this);
        sendCommandAndUpdateStatusBar((PSCommand)cmd);
      } 
    } 
  }
  
  public void commandandReceivedResponse(PSCommand command) {
    if (command instanceof PSCommandFrequencyScan) {
      if (command.getResponse().isSuccess()) {
        setStatusBarCommandSuccess(command);
        setLabelText(this.labelScanStatus, "Running Frequency Scan");
      } else {
        setLabelText(this.labelScanStatus, "Failed to run Frequency Scan");
      } 
    } else if (command instanceof PSCommandRunMode && 
      command.getResponse().isSuccess()) {
      PSCommandRunMode runModeCmd = (PSCommandRunMode)command;
      if (runModeCmd.isSet()) {
        this.devicesController.setRunMode(runModeCmd.getMode());
        setLabelText(this.labelScanStatus, "Stopped Frequency Scan");
      } 
    } 
    if (!command.getResponse().isSuccess()) {
      displayErrorMessage("Error", "Error reading command response");
      setStatusBarCommandFailed(command);
    } 
  }
  
  public void doReloadView() {
    Date lastDataDate = this.frequencyScanManager.getLastUpdate();
    if (lastDataDate != null) {
      boolean update;
      if (this.lastUpdate == null) {
        update = true;
      } else {
        update = (lastDataDate.compareTo(this.lastUpdate) > 0);
      } 
      if (update) {
        this.warfcnVsRssiSeries.getData().clear();
        LinkedHashMap<Short, FrequencyScanSample> samples = this.frequencyScanManager.getSamples();
        for (FrequencyScanSample sample : samples.values()) {
          XYChart.Data<Number, Number> data = new XYChart.Data<>(Short.valueOf(sample.getWarfcn()), Short.valueOf(sample.getRssiMax()));
          data.setNode(new HoveredThresholdNode(sample
                .getWarfcn(), sample.getRssiMax()));
          this.warfcnVsRssiSeries.getData().add(data);
          updateAxisRanges(sample.getWarfcn(), sample.getRssiMax());
        } 
        setAxisRanges();
        Integer expectedSamples = this.frequencyScanManager.getExpectedSamples();
        Integer receivedSamples = this.frequencyScanManager.getReceivedSamples();
        if (expectedSamples != null && receivedSamples != null) {
          ViewUtils.updateProgressBar(this.progressBar, receivedSamples.intValue(), expectedSamples.intValue());
          if (receivedSamples.compareTo(expectedSamples) == 0)
            setLabelText(this.labelScanStatus, "Finished Scanning"); 
        } 
        String error = this.frequencyScanManager.getError();
        if (error != null)
          displayErrorMessage("Error in scan", error); 
        this.lastUpdate = lastDataDate;
      } 
    } 
  }
  
  private LineChart<Number, Number> setUpLineChart(LineChart<Number, Number> lineChart, XYChart.Series<Number, Number> series, String title) {
    if (lineChart != null)
      this.chartsContainer.getChildren().remove(lineChart); 
    lineChart = createChart(series);
    lineChart.setAnimated(false);
    this.chartsContainer.getChildren().add(lineChart);
    return lineChart;
  }
  
  protected LineChart<Number, Number> createChart(XYChart.Series<Number, Number> series) {
    this.yAxis = new NumberAxis();
    this.yAxis.setAutoRanging(false);
    this.yAxis.setLabel("RSSI");
    this.xAxis = new NumberAxis();
    this.xAxis.setLabel("WARFCN");
    this.xAxis.setAutoRanging(false);
    LineChart<Number, Number> lineChart = new LineChart<>(this.xAxis, this.yAxis);
    lineChart.getData().add(series);
    return lineChart;
  }
  
  private void setAxisRanges() {
    this.xAxis.setLowerBound((this.minX - this.step));
    this.xAxis.setUpperBound((this.maxX + this.step));
    this.xAxis.setTickUnit(((this.maxX - this.minX) / 10));
    this.yAxis.setLowerBound(this.minY * 1.1D);
    int rMaxY = this.maxY + 10;
    rMaxY = (rMaxY > 0) ? 0 : rMaxY;
    this.yAxis.setUpperBound(rMaxY);
  }
  
  private void updateAxisRanges(short warfcn, short rssi) {
    this.minX = (this.minX == 50) ? warfcn : ((warfcn < this.minX) ? warfcn : this.minX);
    this.maxX = (warfcn > this.maxX) ? warfcn : this.maxX;
    this.minY = (rssi < this.minY) ? rssi : this.minY;
    this.maxY = (this.maxY == 0) ? rssi : ((rssi > this.maxY) ? rssi : this.maxY);
  }
  
  private void restartAxisRanges() {
    this.minX = 50;
    this.maxX = 4000;
    this.maxY = 0;
    this.minY = 0;
  }
  
  class HoveredThresholdNode extends StackPane {
    HoveredThresholdNode(final int value1, final int value2) {
      setPrefSize(15.0D, 15.0D);
      final int warfcn = value1;
      setOnMouseEntered(new EventHandler<MouseEvent>() {
            public void handle(MouseEvent mouseEvent) {
              FrequencyScannerViewController.this.setLabelText(FrequencyScannerViewController.this.labelCurrentMark, "WARFCN: " + value1 + " RSSI:" + value2);
              FrequencyScannerViewController.HoveredThresholdNode.this.setCursor(Cursor.HAND);
              FrequencyScannerViewController.HoveredThresholdNode.this.toFront();
            }
          });
      setOnMouseExited(new EventHandler<MouseEvent>() {
            public void handle(MouseEvent mouseEvent) {
              FrequencyScannerViewController.HoveredThresholdNode.this.getChildren().clear();
              FrequencyScannerViewController.this.setLabelText(FrequencyScannerViewController.this.labelCurrentMark, "");
              FrequencyScannerViewController.HoveredThresholdNode.this.setCursor(Cursor.CROSSHAIR);
            }
          });
      setOnMouseClicked(new EventHandler<MouseEvent>() {
            public void handle(MouseEvent mouseEvent) {
              FrequencyScannerViewController.this.inputNewWARFCNValue.setText(Integer.toString(warfcn));
            }
          });
    }
  }
}
