package ubiik.wp.view;

import java.util.Arrays;
import java.util.Date;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Optional;
import java.util.Set;
import java.util.function.Function;
import java.util.stream.Collectors;
import javafx.beans.property.SimpleIntegerProperty;
import javafx.beans.value.ObservableValue;
import javafx.collections.FXCollections;
import javafx.collections.ObservableList;
import javafx.event.ActionEvent;
import javafx.fxml.FXML;
import javafx.scene.Node;
import javafx.scene.control.Alert;
import javafx.scene.control.ButtonType;
import javafx.scene.control.CheckBox;
import javafx.scene.control.Label;
import javafx.scene.control.TableCell;
import javafx.scene.control.TableColumn;
import javafx.scene.control.TableView;
import javafx.scene.control.TextInputDialog;
import javafx.util.Callback;
import ubiik.utils.DatatypeConverter;
import ubiik.utils.view.TableUtils;
import ubiik.wp.controller.DevicesController;
import ubiik.wp.controller.MessageController;
import ubiik.wp.model.devices.MulticastGroup;
import ubiik.wp.model.devices.UUEndDevice;
import ubiik.wp.model.messages.pscommands.PSCommand;
import ubiik.wp.model.messages.pscommands.PSCommandAddDevicesToMulticastGroup;
import ubiik.wp.model.messages.pscommands.PSCommandDelegate;
import ubiik.wp.model.messages.pscommands.PSCommandRemoveDevicesFromMulticastGroup;

public class MulticastGroupViewController extends BaseViewController implements PSCommandDelegate {
  private MulticastGroup multicastGroup;
  
  private static final Map<Byte, String> GROUP_STATUS_COLORS = new HashMap<>();
  
  @FXML
  private Label labelGroupId;
  
  static {
    GROUP_STATUS_COLORS.put(Byte.valueOf((byte)0), "#BBBBBB");
    GROUP_STATUS_COLORS.put(Byte.valueOf((byte)1), "#FEDF00");
    GROUP_STATUS_COLORS.put(Byte.valueOf((byte)2), "#90EE90");
    GROUP_STATUS_COLORS.put(Byte.valueOf((byte)3), "#FE7700");
    GROUP_STATUS_COLORS.put(Byte.valueOf((byte)4), "#FE2700");
  }
  
  private final ObservableList<UUEndDevice> data = FXCollections.observableArrayList();
  
  @FXML
  private TableView<UUEndDevice> tableView;
  
  private Set<UUEndDevice> selectedDevices;
  
  public MulticastGroupViewController(DevicesController devicesController, MessageController messageController, CommandStatusReceiver commandStatusReceiver, MulticastGroup multicastGroup) {
    super(devicesController, messageController, commandStatusReceiver);
    this.multicastGroup = multicastGroup;
    this.selectedDevices = new HashSet<>();
  }
  
  public void initialize() {
    this.labelGroupId.setText(Integer.toString(this.multicastGroup.getGroupId()));
    this.tableView.getColumns().clear();
    TableColumn<UUEndDevice, Integer> idxCol = DevicesViewController.createIdxCol(this.tableView, 0.05D);
    TableColumn<UUEndDevice, String> deviceCol = DevicesViewController.createNameCol(this.tableView, 0.3D);
    TableColumn<UUEndDevice, Byte> statusCol = DevicesViewController.createStatusCol(this.tableView, 0.1D);
    TableColumn<UUEndDevice, Long> uplinksCol = DevicesViewController.createUplinksCol(this.tableView, 0.075D);
    TableColumn<UUEndDevice, Date> lastActivityCol = DevicesViewController.createLastActivityCol(this.tableView, 0.125D);
    TableColumn<UUEndDevice, Integer> groupStatus = new TableColumn<>("Status in group");
    groupStatus.prefWidthProperty().bind(this.tableView.widthProperty().multiply(0.2D));
    groupStatus.setSortable(true);
    groupStatus.setCellValueFactory(new Callback<TableColumn.CellDataFeatures<UUEndDevice, Integer>, ObservableValue<Integer>>() {
          public ObservableValue<Integer> call(TableColumn.CellDataFeatures<UUEndDevice, Integer> param) {
            Byte status = (Byte)MulticastGroupViewController.this.multicastGroup.getDevices().get(param.getValue());
            if (status != null)
              return (new SimpleIntegerProperty(status.byteValue())).asObject(); 
            return null;
          }
        });
    groupStatus.setCellFactory(column -> new TableCell<UUEndDevice, Integer>() {
          protected void updateItem(Integer item, boolean empty) {
            super.updateItem(item, empty);
            if (item == null || empty) {
              setText((String)null);
              setStyle("");
            } else {
              setText(MulticastGroup.getNameForStatus(Byte.valueOf(item.byteValue())));
              String color = (String)MulticastGroupViewController.GROUP_STATUS_COLORS.get(Byte.valueOf(item.byteValue()));
              if (color != null) {
                setStyle("-fx-background-color:" + color);
              } else {
                setStyle("");
              } 
            } 
          }
        });
    TableColumn<UUEndDevice, UUEndDevice> selectColumn = TableUtils.column("", javafx.beans.property.ReadOnlyObjectWrapper::new, 60.0D);
    selectColumn.prefWidthProperty().bind(this.tableView.widthProperty().multiply(0.15D));
    selectColumn.setSortable(false);
    selectColumn.setCellFactory(col -> {
          final CheckBox cb = new CheckBox();
          TableCell<UUEndDevice, UUEndDevice> cell = new TableCell<UUEndDevice, UUEndDevice>() {
              public void updateItem(UUEndDevice ed, boolean empty) {
                super.updateItem(ed, empty);
                if (empty) {
                  setGraphic((Node)null);
                } else {
                  cb.setSelected(MulticastGroupViewController.this.selectedDevices.contains(ed));
                  setGraphic(cb);
                } 
              }
            };
          cb.setOnAction(null);
          return cell;
        });
    this.tableView.getColumns().addAll(new TableColumn[] { idxCol, deviceCol, groupStatus, statusCol, lastActivityCol, uplinksCol, selectColumn });
    this.tableView.setItems(this.data);
    reloadView();
  }
  
  public int getMulticastGroupId() {
    return this.multicastGroup.getGroupId();
  }
  
  private List<byte[]> getDevicesInputDialog(String title) {
    TextInputDialog dialog = new TextInputDialog();
    dialog.setTitle(title);
    dialog.setHeaderText("Type devices UUIDs separated by ','");
    List<byte[]> uuids = (List)new LinkedList<>();
    boolean error = false;
    Optional<String> result = dialog.showAndWait();
    if (result.isPresent()) {
      List<String> items = Arrays.asList(((String)result.get()).split("\\s*,\\s*"));
      for (String string : items) {
        try {
          byte[] uuid = DatatypeConverter.parseHexBinary(string);
          if (uuid.length == 16) {
            uuids.add(uuid);
            continue;
          } 
          error = true;
        } catch (Exception e) {
          error = true;
        } 
      } 
    } 
    if (error)
      return null; 
    return uuids;
  }
  
  @FXML
  public void btnAddDevicesPressed() {
    List<byte[]> uuids = getDevicesInputDialog("Add Devices to Multicast Group");
    if (uuids != null) {
      if (uuids.size() > 0) {
        HashMap<Integer, List<byte[]>> groups = new HashMap<>();
        groups.put(Integer.valueOf(0xFFFF & this.multicastGroup.getGroupId()), uuids);
        PSCommandAddDevicesToMulticastGroup cmd = new PSCommandAddDevicesToMulticastGroup(groups, this);
        sendCommandAndUpdateStatusBar((PSCommand)cmd);
      } 
    } else {
      displayErrorMessage("Format error", "Devices UUIDs are 32 HEX character long separated by ','");
    } 
  }
  
  @FXML
  public void btnRemoveDevicesPressed() {
    List<byte[]> uuids = getDevicesInputDialog("Remove Devices from Multicast Group");
    if (uuids != null) {
      if (uuids.size() > 0) {
        HashMap<Integer, List<byte[]>> groups = new HashMap<>();
        groups.put(Integer.valueOf(0xFFFF & this.multicastGroup.getGroupId()), uuids);
        PSCommandRemoveDevicesFromMulticastGroup cmd = new PSCommandRemoveDevicesFromMulticastGroup(groups, this);
        sendCommandAndUpdateStatusBar((PSCommand)cmd);
      } 
    } else {
      displayErrorMessage("Format error", "Devices UUIDs are 32 HEX character long separated by ','");
    } 
  }
  
  public void commandandReceivedResponse(PSCommand command) {
    if (command.getResponse().isSuccess()) {
      setStatusBarCommandSuccess(command);
    } else {
      displayErrorMessage("Error", "The command failed");
      setStatusBarCommandFailed(command);
    } 
  }
  
  private synchronized void endDeviceItemSelected(UUEndDevice ed, boolean selected) {
    if (selected) {
      this.selectedDevices.add(ed);
    } else {
      this.selectedDevices.remove(ed);
    } 
  }
  
  public synchronized void doReloadView() {
    UUEndDevice selectedDevice = this.tableView.getSelectionModel().getSelectedItem();
    this.data.clear();
    Set<UUEndDevice> endDevices = this.multicastGroup.getDevices().keySet();
    this.data.addAll(endDevices);
    this.selectedDevices = (Set<UUEndDevice>)this.selectedDevices.stream().filter(ed -> endDevices.contains(ed)).collect(Collectors.toSet());
    this.tableView.sort();
    this.tableView.refresh();
    this.tableView.getSelectionModel().select(selectedDevice);
  }
  
  public void btnRemoveSelectedDevicesPressed() {
    if (this.selectedDevices.size() > 0) {
      Alert alert = new Alert(Alert.AlertType.CONFIRMATION);
      alert.setTitle("Remove Devices");
      alert.setHeaderText("You are about to remove " + this.selectedDevices.size() + " devices from this group");
      alert.setContentText("Continue?");
      Optional<ButtonType> result = alert.showAndWait();
      if (result.isPresent() && result.get() == ButtonType.OK)
        removeSelectedDevices(); 
    } else {
      displayErrorMessage("No devices selected", "Please select at least one device to remove");
    } 
  }
  
  private synchronized void removeSelectedDevices() {
    Map<Integer, List<byte[]>> multicastGroups = new HashMap<>();
    List<byte[]> devices = (List<byte[]>)this.selectedDevices.stream().map(ed -> ed.getUuid()).collect(Collectors.toList());
    multicastGroups.put(Integer.valueOf(this.multicastGroup.getGroupId()), devices);
    PSCommandRemoveDevicesFromMulticastGroup cmd = new PSCommandRemoveDevicesFromMulticastGroup(multicastGroups, this);
    sendCommandAndUpdateStatusBar((PSCommand)cmd);
  }
}
