package ubiik.wp.view;

import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.Optional;
import java.util.function.Function;
import java.util.stream.Collectors;
import javafx.beans.property.ReadOnlyObjectWrapper;
import javafx.beans.value.ObservableValue;
import javafx.collections.FXCollections;
import javafx.collections.ObservableList;
import javafx.event.ActionEvent;
import javafx.fxml.FXML;
import javafx.scene.control.Button;
import javafx.scene.control.ChoiceDialog;
import javafx.scene.control.TableCell;
import javafx.scene.control.TableColumn;
import javafx.scene.control.TableView;
import javafx.scene.control.TextField;
import javafx.scene.control.cell.PropertyValueFactory;
import javafx.stage.FileChooser;
import javafx.stage.Stage;
import javafx.util.Callback;
import ubiik.utils.view.TableUtils;
import ubiik.wp.controller.DevicesController;
import ubiik.wp.controller.MessageController;
import ubiik.wp.model.devices.MulticastGroup;
import ubiik.wp.model.devices.UUEndDevice;
import ubiik.wp.model.messages.pscommands.PSCommand;
import ubiik.wp.model.messages.pscommands.PSCommandAddDevicesToMulticastGroup;
import ubiik.wp.model.messages.pscommands.PSCommandDelegate;
import ubiik.wp.model.messages.pscommands.PSCommandSetMulticastGroups;
import ubiik.wp.persistence.MulticastGroupJSONParser;

public class MulticastGroupsViewController extends BaseViewController implements PSCommandDelegate {
  @FXML
  private TextField inputNewMulticastGroup;
  
  @FXML
  private TableView<MulticastGroup> tableView;
  
  private MainViewController mainViewController;
  
  private final ObservableList<MulticastGroup> data = FXCollections.observableArrayList();
  
  private Stage mainStage;
  
  private static final String IMPORT_CHOICE_OVERRIDE_ALL = "Override all groups and devices with the contents of this file";
  
  private static final String IMPORT_CHOICE_OVERRIDE_GROUPS = "Override only groups in this file (keep other groups not in the file as they are)";
  
  private static final String IMPORT_CHOICE_ADD = "Add devices to existing groups and create new groups if necessary";
  
  public MulticastGroupsViewController(Stage mainStage, DevicesController devicesController, MessageController messageController, CommandStatusReceiver commandStatusReceiver, MainViewController mainViewController) {
    super(devicesController, messageController, commandStatusReceiver);
    this.mainViewController = mainViewController;
    this.mainStage = mainStage;
  }
  
  public void initialize() {
    this.tableView.getColumns().clear();
    TableColumn<MulticastGroup, Integer> idCol = new TableColumn<>("Group ID");
    idCol.prefWidthProperty().bind(this.tableView.widthProperty().multiply(0.1D));
    idCol.setSortable(true);
    idCol.setCellValueFactory(new PropertyValueFactory<>("groupId"));
    TableColumn<MulticastGroup, Integer> countCol = new TableColumn<>("# Devices");
    countCol.prefWidthProperty().bind(this.tableView.widthProperty().multiply(0.1D));
    countCol.setSortable(true);
    countCol.setCellValueFactory(new PropertyValueFactory<>("totalDevices"));
    TableColumn<MulticastGroup, MulticastGroup> editColumn = TableUtils.column("", ReadOnlyObjectWrapper::new, 60.0D);
    editColumn.prefWidthProperty().bind(this.tableView.widthProperty().multiply(0.15D));
    editColumn.setSortable(false);
    editColumn.setCellFactory(col -> {
          final Button editButton = new Button("Details");
          TableCell<MulticastGroup, MulticastGroup> cell = new TableCell<MulticastGroup, MulticastGroup>() {
              public void updateItem(MulticastGroup group, boolean empty) {
                super.updateItem(group, empty);
                if (empty) {
                  setGraphic(null);
                } else {
                  setGraphic(editButton);
                } 
              }
            };
          editButton.setOnAction(e -> {
        	  showDetails(cell.getTableRow().getItem());
          });
          return cell;
        });
    TableColumn<MulticastGroup, Long> joinedCol = createStatusCountCol("Joined", this.tableView, 0.1D, (byte)2);
    TableColumn<MulticastGroup, Long> joiningCol = createStatusCountCol("Joining", this.tableView, 0.1D, (byte)1);
    TableColumn<MulticastGroup, Long> leavingCol = createStatusCountCol("Leaving", this.tableView, 0.1D, (byte)3);
    TableColumn<MulticastGroup, Long> unknownCol = createStatusCountCol("Unknown", this.tableView, 0.1D, (byte)0);
    this.tableView.getColumns().addAll(new TableColumn[] { idCol, countCol, joinedCol, joiningCol, leavingCol, unknownCol, editColumn });
    this.tableView.setItems(this.data);
  }
  
  public void doReloadView() {
    this.data.clear();
    this.data.addAll(this.devicesController.getMulticastGroups().values());
    MulticastGroup selectedGroup = this.tableView.getSelectionModel().getSelectedItem();
    this.tableView.refresh();
    this.tableView.getSelectionModel().select(selectedGroup);
  }
  
  @FXML
  public void btnAddMulticastGroupPressed() {
    int groupId = -1;
    boolean error = false;
    try {
      groupId = Integer.parseInt(this.inputNewMulticastGroup.getText());
      if (groupId < 1 || groupId > 65535)
        error = true; 
      if (this.devicesController.mutlicastGroupExists(groupId))
        error = true; 
    } catch (Exception e) {
      error = true;
    } 
    if (!error && groupId >= 0) {
      MulticastGroup mcg = new MulticastGroup(groupId);
      error = !this.devicesController.addMulticastGroup(mcg);
    } 
    setTextFieldHasError(this.inputNewMulticastGroup, error);
  }
  
  public void btnImportMulticastGroupsPressed() {
    FileChooser fileChooser = new FileChooser();
    fileChooser.setTitle("Open Multicast Groups File");
    File selectedFile = fileChooser.showOpenDialog(this.mainStage);
    if (selectedFile != null)
      try {
        InputStream is = new FileInputStream(selectedFile);
        Map<Integer, List<byte[]>> groups = MulticastGroupJSONParser.parseJSON(new InputStreamReader(is));
        is.close();
        String choiceOverride = "Override all groups and devices with the contents of this file";
        String choiceOverrideGroups = "Override only groups in this file (keep other groups not in the file as they are)";
        String choiceAdd = "Add devices to existing groups and create new groups if necessary";
        List<String> choices = new ArrayList<>();
        choices.add(choiceOverride);
        choices.add(choiceOverrideGroups);
        choices.add(choiceAdd);
        ChoiceDialog<String> dialog = new ChoiceDialog<>(choiceAdd, choices);
        dialog.setTitle("Import Multicast Groups");
        dialog.setHeaderText("Choose how to import multicast groups");
        dialog.setContentText("The file contains " + groups.size() + " multicast groups. Please select how to import them:\n");
        Optional<String> result = dialog.showAndWait();
        if (result.isPresent()) {
          PSCommandSetMulticastGroups pSCommandSetMulticastGroups = null;
          PSCommand cmd = null;
          if (((String)result.get()).equals(choiceOverride)) {
            pSCommandSetMulticastGroups = new PSCommandSetMulticastGroups(groups, this);
          } else if (((String)result.get()).equals(choiceAdd)) {
            PSCommandAddDevicesToMulticastGroup pSCommandAddDevicesToMulticastGroup = new PSCommandAddDevicesToMulticastGroup(groups, this);
          } else if (((String)result.get()).equals(choiceOverrideGroups)) {
            Map<Integer, MulticastGroup> currentGroups = this.devicesController.getMulticastGroups();
            for (Map.Entry<Integer, MulticastGroup> entry : currentGroups.entrySet()) {
              if (!groups.containsKey(entry.getKey())) {
                List<byte[]> groupDevices = (List<byte[]>)((MulticastGroup)entry.getValue()).getDevices().keySet().stream().map(dev -> dev.getUuid()).collect(Collectors.toList());
                groups.put(entry.getKey(), groupDevices);
              } 
            } 
            pSCommandSetMulticastGroups = new PSCommandSetMulticastGroups(groups, this);
          } 
          if (pSCommandSetMulticastGroups != null)
            sendCommandAndUpdateStatusBar((PSCommand)pSCommandSetMulticastGroups); 
        } 
      } catch (IOException e) {
        e.printStackTrace();
      }  
  }
  
  public void btnExportMulticastGroupsPressed() {
    FileChooser fileChooser = new FileChooser();
    File file = fileChooser.showSaveDialog(this.mainStage);
    if (file != null)
      try {
        MulticastGroupJSONParser.multicastGroupsToFile(this.devicesController
            .getMulticastGroups().values(), file
            .getAbsolutePath());
      } catch (IOException e) {
        displayErrorMessage("Error saving file", e.getMessage());
      }  
  }
  
  public void btnRefreshPressed() {
    reloadView();
  }
  
  public void showDetails(MulticastGroup mcg) {
    if (mcg != null)
      this.mainViewController.displayMulticastGroupView(mcg); 
  }
  
  public void commandandReceivedResponse(PSCommand command) {
    if (command.getResponse().isSuccess()) {
      setStatusBarCommandSuccess(command);
    } else {
      displayErrorMessage("Error", "The command failed");
      setStatusBarCommandFailed(command);
    } 
  }
  
  public static TableColumn<MulticastGroup, Long> createStatusCountCol(String name, TableView<MulticastGroup> tableView, double widthProperty, final byte status) {
    TableColumn<MulticastGroup, Long> statusCol = new TableColumn<>(name);
    statusCol.prefWidthProperty().bind(tableView.widthProperty().multiply(widthProperty));
    statusCol.setSortable(true);
    statusCol.setCellValueFactory(new Callback<TableColumn.CellDataFeatures<MulticastGroup, Long>, ObservableValue<Long>>() {
          public ObservableValue<Long> call(TableColumn.CellDataFeatures<MulticastGroup, Long> p) {
            return new ReadOnlyObjectWrapper<>(Long.valueOf(((MulticastGroup)p.getValue()).getStatusCount(status)));
          }
        });
    return statusCol;
  }
}
